/**
 * Human Mate Choice is a Complex System
 * 
 * Copyright 2008 by Paul E. Smaldino and Jeffrey C. Schank
 * This is a simple model of human mate choice in which agents are move in space and can only interact with close neighbors.
 * 
 * This model is fully described and analyzed in: 
 * Smaldino PE & Schank JC. (2012). Human mate choice is a complex system. Complexity, 17(5), 11-22.
 * 
 * Our model is a dramatic extension of Kalick and Hamilton's original mate choice model:
 * Kalick SM & Hamilton TE. (1986). The matching hypothesis reconsidered. Journal of Personality and Social Psychology, 51, 673-682.
 * 
 * Requires the MASON simulation library, which is in the project's build path. Mason is available here:
 * http://cs.gmu.edu/~eclab/projects/mason/
 */

package agents;

import agents.states.SimStateWithSimController;
import sim.field.grid.SparseGrid2D; 
import sim.portrayal.grid.SparseGridPortrayal2D;
import sim.util.Bag;
import sim.util.Int2D;
import java.util.Random;
 
public class AgentsSimulation extends SimStateWithSimController { //extend the class SimState

        public SparseGrid2D agentsSpace; // 2D space for agents
        public SparseGridPortrayal2D agentsPortrayal = null;
        public int gridWidth = 200; // width of the space
        public int gridHeight = 200; // Height of the space
        public boolean uniqueLocation = false;
        public boolean do_not_stack = true;
        public boolean moveRandomly = true;
        public boolean aggregate = false;
        public boolean align = false;
        public int searchRadius = 10;  // we should be able to change this in the model tab
        public double majority = .5; // we should be able to change this
        public int density = 3; // we should be able to change this
        
        public double[][] movementProbs = {{1.0,0.0,1.0},{0.0,0.0,0.0},{0.0,0.0,0.0}};
        public double[][] zigzag = {{1.0,0.0,1.0},{0.0,0.0,0.0},{0.0,0.0,0.0}};
        public double[][] brownian = {{1.0,1.0,1.0},{1.0,0.0,1.0},{1.0,1.0,1.0}};
        
        
        public boolean probablistic = true;
        public double[][] movementProbsSums = null;
        public int[][] orientationLookUp = null;
        public Int2D[][] directionLookUp = null;
        public Probabilities probs = new Probabilities();
        public Int2D[] randomMovement = null;
        
        /*
        * Data for the matching simulation
        */

           public int males = 1000; // the numbers used in the original
           public int females = 1000;
           public double exponent = 3;
           public double maxAttractiveness = 10;
           public int max_d = 51; // maximum number of dates to no choosiness
           public int divisor = 50;
           public int matchRadius = 5;
           public final static int ATTRACTIVE = 0;
           public final static int SIMILAR = 1;
           public final static int MIXED = 2;
           public final static int MIXED2 = 3;
           public String[] rules = {"attractive", "similar", "mixed"};
           public int rule = ATTRACTIVE;  // by default
          
           public double p = .5; //probabilistically use both Rules 1 and 2
           
           public boolean maxMatchRadius = true; // increases processing speed when there is no limit to the search radius.
           public boolean prettierAtClosingTime = true; // 
           public boolean replacement = false; // replace matched agents with new agents? 
           public int[] numberMated = new int[10];
           public boolean printOnlyEndResult = true; // prints only the final stats to (correl) file.
           public int simNumber = 0;
           
           public boolean normalDist = false;
           public double normalSD = 1.8; //Aron: 1.2. K&H: 1.8.
           public boolean roundNormalInt = true; //if using a normal distr, do we round to nearest integer? 
           
           
           public int[] marriageAge = null; // this will allow a histogram of age at marriage
           public int[][]matePairs = new int[(int)maxAttractiveness][(int)maxAttractiveness];
           				//this will allow us to look at how well matched everyone is.
           public int[][] marriageAgeIndiv = new int[(int)maxAttractiveness]
                                                     [(int)(6*(double)(males)/(maxAttractiveness))];
           				
           public int[] marriageAgeDates = null; // this will allow a histogram of age at marriage
           				//this will allow us to look at how well matched everyone is.
           public int[][] marriageAgeDatesIndiv = new int[(int)maxAttractiveness]
                                                     [(int)(6*(double)(males)/(maxAttractiveness))];
        
           public int[] marriageAgeIndivCounter = new int[(int)maxAttractiveness];
          
           public double[] meanMarriageAgeTime;
           public double[] meanMarriageAgeDates;
           public int matchCounter;
           
           
           public boolean useSpace = true;
           public int moveProb = 0; //0 = Zigzag, 1 = Brownian
           
           //stuff for randomDates
           public boolean randomDates = false; //is starting value of d randomly determined?
           public double rdMean = 10;
           public double rdSD = 10;
           public double rdMin = -10;
           public double rdMax = 50;
           
           
           //variables for batch runs
           public Bag recycledAgents = new Bag();
           public int simCount = 0;
           public String dataFileName = "data.txt";
           public String dataFolderName = "data";
           public boolean saveDataToFile = false;
           public boolean writeDataToConsole = true;
           
           public boolean printedInfo = false;
           
        public AgentsSimulation(long seed) {
             super(seed); //use the constructor for SimState
        }
        
        
//        public boolean getautoSim(){return autoSimulation;}
//    	public void setautoSim(boolean b){autoSimulation = b;}
        
      public String getInitialization(){ return "Initialization Parameters";}
        public int getgridWidth (){    	return gridWidth;        }
        public void setgridWidth(int w){    if(w > 0)	gridWidth = w;        }
        public int getgridHeight(){        	return gridHeight; }
        public void setgridHeight(int h){        if(h > 0) gridHeight = h;    }   
        public int getfemales(){            return females;        }       
        public void setfemales(int x){            if(x > 0)    females = x;       }       
        public int getmales(){            return males;       }        
        public void setmales(int x){            if(x > 0)        males = x;        }
        public double getmaxAttractiveness(){            return maxAttractiveness;    }       
        public void setmaxAttractiveness(double x){  if(x > 0)     maxAttractiveness = x; }
        public boolean getreplacement(){        	return replacement;       }       
        public void setreplacement(boolean b){        	replacement = b;       }   
        public boolean getrandomDates(){        	return randomDates;       }       
        public void setrandomDates(boolean b){        	randomDates = b;       }   
        
      public String getMateChoice(){ return "Mate Choice Parameters";}
        public int getrule(){        	return rule;      }       
        public void setrule(int x){        	if(x >= 0)       		rule = x;        }     
        public Object domrule() 
        { 
        return new String[] { "Most Attractive", "Similarity", "Mixed-Average", "Mixed-Probabilistic" };
        }
        public double getexponent(){            return exponent;        }       
        public void setexponent(double x){           if(x > 0)   exponent = x;        }       

        public int getmaxDates(){   return max_d;        }        
        public void setmaxDates(int x){  if(x > 1)    {  max_d = x;  divisor = x - 1;  }    }  
        public boolean getprettierAtClosingTime(){        	return prettierAtClosingTime;       }       
        public void setprettierAtClosingTime(boolean b){        	prettierAtClosingTime = b;       }  
        public int getmatchRadius(){            return matchRadius;        }        
        public void setmatchRadius(int x){            if(x > 0)   matchRadius = x;    }        
        public boolean getmaxMatchRadius(){        	return maxMatchRadius;       }       
        public void setmaxMatchRadius(boolean b){        	maxMatchRadius = b;       }  
        
      public String getNormalDist(){ return "Normal Distribution of Attractiveness";}
      	public boolean getnormalDist(){            return normalDist;        }        
      	public void setnormalDist(boolean b){    normalDist = b;        }
      	public double getnormalSD(){            return normalSD;    }       
      	public void setnormalSD(double x){  if(x > 0)     normalSD = x; }
      	public boolean getroundNormalInt(){            return roundNormalInt;        }        
      	public void setroundNormalInt(boolean b){    roundNormalInt = b;        }    
      	
     public String getMovement(){ return "Movement Parameters";}
        public int getmoveProb(){	return moveProb;}
        public void setmoveProb(int b){ 
        	moveProb = b;
        	switch(b) {
        	case 0: movementProbs = zigzag; break;
        	case 1: movementProbs = brownian; break;
        	default: movementProbs = zigzag;
        	}
        }
        public Object dommoveProb() 
        { 
        return new String[] { "Zigzag", "Brownian" };
        }
 
       
        public void setMovementRule(int moveProb){
        	switch(moveProb) {
        		case 0: movementProbs = zigzag; break;
        		case 1: movementProbs = brownian; break;
        		default: movementProbs = zigzag;
        	}
        }
        
      public String getDataExport(){ return "Data Export Parameters";}
        public boolean getwriteDataToConsole(){            return writeDataToConsole;        }        
        public void setwriteDataToConsole(boolean b){    writeDataToConsole = b;        }
        public boolean getsaveDataToFile(){            return saveDataToFile;        }        
        public void setsaveDataToFile(boolean b){    saveDataToFile = b;        }
        
        public boolean getprintOnlyEndResult(){            return printOnlyEndResult;        }        
        public void setprintOnlyEndResult(boolean b){    printOnlyEndResult = b;        }
        public String getdataFileName(){            return dataFileName;       }        
        public void setdataFileName(String s){  if(s.length() > 0)   dataFileName = s;     }       
        public String getdataFolderName(){            return dataFolderName;   }
        public void setdataFolderName(String s){   if(s.length() > 0)     dataFolderName = s; }

        

        
        /**
         * Returns a random double from a random distribution with
         * mean = 5.5 and SD = normalSD. If roundNormalInt is true, it returns the nearest 
         * integer value. All values are capped between 1 and 10.  
         */
        public double normalAttr(){
        	double d = random.nextGaussian();
        	d = d*normalSD + maxAttractiveness/2 + 0.5;
        	if(d < 1)
        		d = 1;
        	else if(d > maxAttractiveness)
        		d = maxAttractiveness;
        	if(roundNormalInt)
        		d = Math.rint(d);
        	
        	return d;
        }
        
        
        public void intializeDataArrays(){
        	matePairs = new int[(int)maxAttractiveness][(int)maxAttractiveness];
				//this will allow us to look at how well matched everyone is.
        	marriageAgeIndiv = new int[(int)maxAttractiveness]
                                         [(int)(5000*(double)(males)/(maxAttractiveness))];
        	marriageAgeDatesIndiv = new int[(int)maxAttractiveness]
                                       [(int)(5000*(double)(males)/(maxAttractiveness))];
				//should be plenty big.
        	marriageAgeIndivCounter = new int[(int)maxAttractiveness];
        }
        
        
        public void initializeMarriageAge(){
        	if(prettierAtClosingTime && maxMatchRadius && !replacement){
        		marriageAge = new int[max_d + 100];
        		marriageAgeDates = new int[max_d + 100];
        	}
        	else{
        		marriageAge = new int[1000];//if CT = false, record all over 1,000 as 1,000. this is a q&d fix.
        		marriageAgeDates = new int[1000];
        	}
        }
        
        public int initD(){
        	if(!randomDates)
        		return 1;
        	else{
        		double a = rdMean + random.nextGaussian()*rdSD;
        		if(a < rdMin) 
        			a = rdMin;
        		else if(a > rdMax)
        			a = rdMax;
        		return (int)a;
        	}
        }
        
        
   public void start(){
	   //stuff for batch runs
        super.start(); // reuse the SimState start method
   		simCount++;//added, probably unnecessary --PS 1/13/10
   		checkReleaseAutomatedControl();
   		sc = checkAutomatedControl(); // allow the observer and controller to interact
   		checkReleaseAutomatedControl();
   		recycledAgents.clear();

   		//begin model-relevant instructions
           simNumber++;
           agentsSpace = new SparseGrid2D(gridWidth, gridHeight); 
           printedInfo = false;
                                            
           
           matchCounter = 0;
           if(!replacement){
        	   meanMarriageAgeTime = new double[males + females];
        	   meanMarriageAgeDates = new double[males + females];
           }
           else{
        	   meanMarriageAgeTime = new double[1000000];
        	   meanMarriageAgeDates = new double[1000000];
           }
           
           setMovementRule(moveProb);
           
           Probabilities probs = new Probabilities();
           randomMovement = probs.makeRandomMovement();
           movementProbsSums = probs.makeMovementProbsSum(movementProbs);
           directionLookUp = probs.makeDirectionLookUp();
           orientationLookUp = probs.makeOrientationLookUp();
           
           initializeMarriageAge();
           intializeDataArrays();
           new Observer(this, dataFileName , dataFolderName , saveDataToFile, sc, precision);
           
           numberMated = new int[10];
                /*
                 * Now create the agents and put them in the space.
                 */ 

            for(int i=0;i<females;i++){ // 0 to the number of agents - 1
               int x = random.nextInt(gridWidth); // a random number <
                                                  //gridWidth
               int y = random.nextInt(gridHeight); //a random number <
                                                  //gridHeight
               Int2D dir = randomMovement[random.nextInt(8)];
               Person a = new Person(x,y,dir.x,dir.y);
               a.sex = Person.FEMALE;
               a.d = initD();
               a.makePortrayl((float)1.0, (float)0.08 ,(float)0.57, 1);   // define a color 
           
               if(normalDist)
            	   a.attractiveness = normalAttr();
               else
            	   a.attractiveness = (double)this.random.nextInt((int)this.maxAttractiveness)+1;  
               
               a.movementProbs = movementProbsSums;
          	   a.orientationLookUp = orientationLookUp;
          	   a.directionLookUp = directionLookUp;
          	   a.randomMovement = randomMovement;         	   
          	   a.orientation = a.lookUpOrientation(dir.x,dir.y);
               agentsSpace.setObjectLocation(a, x,y); //put it at location x, y
               a.stop = schedule.scheduleRepeating(a); // schedule it to repeat in
                                              // in the simulation
                }
            
            for(int i=0;i<males;i++){ // 0 to the number of agents - 1
                int x = random.nextInt(gridWidth); // a random number <
                                                   //gridWidth
                int y = random.nextInt(gridHeight); //a random number <
                                                   //gridHeight
                Int2D dir = randomMovement[random.nextInt(8)];
                Person a = new Person(x,y,dir.x,dir.y);
                a.sex = Person.MALE;
                a.d = initD();
                a.makePortrayl(0, 0,1, 1);   // define a color 
                if(normalDist)
             	   a.attractiveness = normalAttr();
                else
             	   a.attractiveness = (double)this.random.nextInt((int)this.maxAttractiveness)+1;
                a.movementProbs = movementProbsSums;
           	   a.orientationLookUp = orientationLookUp;
           	   a.directionLookUp = directionLookUp;
           	   a.randomMovement = randomMovement;         	   
           	   a.orientation = a.lookUpOrientation(dir.x,dir.y);
                agentsSpace.setObjectLocation(a, x,y); //put it at location x, y
                a.stop = schedule.scheduleRepeating(a); // schedule it to repeat
                 }
        }//end start methods
   
   
        public static void main(String[] args) {
            doLoop(AgentsSimulation.class, args); 
            System.exit(0);//Stop the program when finished.
        }


}//end class